#include "globals.h"
#include "IterativeAssembler.h"
#include "Kmer.h"
#include "Contig.h"
#include "Log.h"

#include <iostream>
#include <algorithm>
#include <cstdio>
#include <vector>
#include <cmath>

using namespace std;

void IterativeAssembler::Assemble(vector<Contig> &contigs)
{
    FastAReader contig_reader(option.graphfile);

    Contig contig;
    string comment;
    tmp_contigs.resize(contig_reader.NumReads());
    cout << tmp_contigs.size() << endl;
    int i = 0;
    while (contig_reader.Read(contig, comment))
    {
        int index = comment.size()-1;
        while (isdigit(comment[index]))
            --index;
        contig.sum_coverage = atoi(comment.c_str() + index + 1);
        contig.Encode();
        //tmp_contigs.push_back(contig);
        tmp_contigs[i++] = contig;
    }


    Iterate();

    //contigs = tmp_contigs;
    contigs.swap(tmp_contigs);

    kmerLength = option.maxk;

    HashGraph *hash_graph = new HashGraph();
#pragma omp parallel for
    for (int i = 0; i < (int)contigs.size(); ++i)
        hash_graph->InsertSequence(contigs[i]);

    RewindShortReads();
    while (true)
    {
        vector<Read> &reads = GetShortReads();
        if (reads.size() == 0)
            break;

#pragma omp parallel for
        for (int i = 0; i < (int64)reads.size(); ++i)
        {
            if (!reads[i].IsActive() || reads[i].Size() < kmerLength)
                continue;

            Sequence seq;
            seq = reads[i];
            seq.Trim(option.trim);
            hash_graph->AddEdgesFromSequence(seq);
        }
    }


    vector<Sequence> &long_reads = GetLongReads();
#pragma omp parallel for
    for (int i = 0; i < (int64)long_reads.size(); ++i)
    {
        hash_graph->AddEdgesFromSequence(long_reads[i]);
    }

    hash_graph->RefreshEdges();
    int deadend = hash_graph->RemoveDeadEnd(kmerLength * 2);
    LogMessage("remove %d deadend\n", deadend);
    double avg = hash_graph->AverageCoverage();
    LogMessage("average coverage %.4f", avg);
//    int low_cov = hash_graph->RemoveLowCoverageContigs(sqrt(avg));
//    LogMessage("remove %d low coverage contigs\n", low_cov);
    int bubble = hash_graph->RemoveBubble();
    LogMessage("remove %d bubbles\n", bubble);
    hash_graph->Assemble(contigs);
}

void IterativeAssembler::Iterate()
{
    cout << "size of kmer node " << sizeof(KmerNode) << endl; 
    cout << "kmer size " << KMERSIZE << endl;
    cout << "size of kmer " << sizeof(Kmer) << endl;
    vector<Kmer> branches;

    //while (true);

    kmerLength = option.mink;
    LogMessage("start iteration from mink = %d to maxk = %d\n", kmerLength, option.maxk);
    while (kmerLength <= option.maxk)
    {
        //int x;
        cout << "before initialize" << endl;
        //cin >> x;
        //cout << "start initialize" << endl;
        contig_graph.Initialize(tmp_contigs, branches);

        //while (true);

        cout << "after initialize" << endl;
        //cin >> x;

        RewindShortReads();
        vector<Read> &reads = GetShortReads();
#pragma omp parallel for
        for (int64 i = 0; i < (int64)reads.size(); ++i)
        {
            if (reads[i].IsActive() || reads[i].Size() < kmerLength)
            {
                Sequence seq;
                seq = reads[i];
                seq.Trim(option.trim);
                if (!contig_graph.BuildEdgesFromSequence(seq))
                    reads[i].Inactivate();
            }
        }

        cout << "after build edges" << endl;

        vector<Sequence> &long_reads = GetLongReads();
        cout << "long read " << long_reads.size() << endl;
#pragma omp parallel for
        for (int i = 0; i < (int)long_reads.size(); ++i)
            contig_graph.BuildEdgesFromSequence(long_reads[i]);

        cout << "before refresh" << endl;

        contig_graph.Refresh();

        cout << "after refresh" << endl;

        int minLength = kmerLength * 2;
        if (minLength < 75)
            minLength = 75;

        int deadend = 0;
        //if (kmerLength < option.maxk)
        {
//            if (kmerLength == option.mink)
//            {
//                option.cover = contig_graph.AverageCoverage();
//                option.cover = sqrt(option.cover);
//                if (option.cover < 2)
//                    option.cover = 2;
//
//                cout << option.cover << endl;
//                //option.cover = 2;
//
//                for (int i = 3; i >= 0; --i)
//                {
//                    int64 trimmed = contig_graph.RemoveDeadEnd(minLength >> i);
//                    LogMessage("remove %d deadend\n", trimmed);
//                    contig_graph.MergeContigs();
//                    LogMessage("merge contigs\n");
//                }
//
//                int64 low_coverage = contig_graph.RemoveLowCoverageContigs(option.cover);
//                LogMessage("remove %d low coverage contigs\n", low_coverage);
//
//                contig_graph.MergeContigs();
//            }
//
        }

        //if (kmerLength < option.maxk)
            deadend = contig_graph.RemoveDeadEnd(minLength);

        if (kmerLength == option.maxk)
        {
            int bubbles = contig_graph.RemoveBubble();
            LogMessage("remove %d bubbles\n", bubbles);
        }

        //cout << "after remove deadend or bubbles" << endl;

        contig_graph.MergeContigs();

        for (unsigned i = 0; i < tmp_contigs.size(); ++i)
        {
            tmp_contigs[i].Dispose();
        }

        vector<Contig>().swap(tmp_contigs);
        int num = contig_graph.Assemble(tmp_contigs, branches);
        LogMessage("k = %d, remove %d deadend, remain %d branches\n", kmerLength, deadend, num);

        ++kmerLength;
    }

    contig_graph.Clear();
}
