#include "globals.h"
#include "Log.h"
#include "Utils.h"
#include "Sequence.h"
//#include "HashGraph.h"
//#include "Contig.h"

#include <cstdio>
#include <cstring>
#include <cctype>
#include <algorithm>
#include <iostream>
#include <map>
#include <string>
#include <sstream>
#include <iomanip>

#include <getopt.h>

using namespace std;

struct Option
{
    string long_name;
    string short_name;
    void *pointer;
    OptionType type;
    string description;
    string default_value;
    string value;

    Option(const string &long_name, const string &short_name,
            void *pointer, OptionType type, const string &description, 
            const string &default_value)
    {
        if (short_name.size() > 1 || long_name.size() < 2)
        {
            fprintf(stderr, "invalid parameter: %s, %s\n", 
                    long_name.c_str(), short_name.c_str());
            throw exception();
        }

        this->long_name = long_name;
        this->short_name = short_name;
        this->pointer = pointer;
        this->type = type;
        this->description = description;
        this->default_value = default_value;
    }

    void Process()
    {
        if (value == "")
            value = default_value;

        if (value != "")
        {
            switch (type)
            {
                case OptionBool:
                    if (value != "" || value == "false")
                        *(bool *)pointer = true;
                    break;

                case OptionInt:
                    *(int *)pointer = atoi(value.c_str());
                    break;

                case OptionDouble:
                    *(double *)pointer = atof(value.c_str());
                    break;

                case OptionString:
                    *(string *)pointer = value;
            }
        }
    }

    string ToString()
    {
        string s;

        if (short_name != "")
            s += "  -" + short_name + ", ";
        else
            s += "      ";

        s += "--" + long_name;
        if (type != OptionBool)
            s += " arg";

        if (default_value != "")
            s += " (=" + default_value + ")";

        while (s.size() < 30)
            s += " ";

        s += " " + description;

        return s;
    }
};

static const int FastaMaxLine = 80;

static map<string, void *> parameters;
static map<string, ParameterType> types;
static vector<Option> options;

void AddOption(const string &long_name, const string &short_name,
        void *pointer, OptionType type, const string &description, const string &default_value)
{
    for (unsigned i = 0; i < options.size(); ++i)
    {
        if (options[i].long_name == long_name
                || (short_name != "" && options[i].short_name == short_name))
        {
            fprintf(stderr, "parameter already exist: %s, %s\n",
                    long_name.c_str(), short_name.c_str());
            throw exception();
        }
    }

    options.push_back(Option(long_name, short_name, pointer, type, description, default_value));
}

void AddOption(const std::string &long_name, const std::string &short_name,
        bool &bool_option, const std::string &description)
{
    AddOption(long_name, short_name, &bool_option, OptionBool, description, "");
}

void AddOption(const std::string &long_name, const std::string &short_name,
        int &int_option, const std::string &description)
{
    stringstream ss;
    ss << int_option;
    AddOption(long_name, short_name, &int_option, OptionInt, description, ss.str());
}

void AddOption(const std::string &long_name, const std::string &short_name,
        double &double_option, const std::string &description)
{
    stringstream ss;
    ss << double_option;
    AddOption(long_name, short_name, &double_option, OptionDouble, description, ss.str());
}

void AddOption(const std::string &long_name, const std::string &short_name,
        std::string &string_option, const std::string &description)
{
    AddOption(long_name, short_name, &string_option, OptionString, description, string_option);
}


void ProcessOptions(int &argc, char *argv[])
{
    struct option long_options[options.size()+1];
    string short_options;
    for (unsigned i = 0; i < options.size(); ++i)
    {
        long_options[i].name = options[i].long_name.c_str();
        long_options[i].flag = 0;
        long_options[i].val = 
            (options[i].short_name != "" ? options[i].short_name[0] : 0);

        if (options[i].type == OptionBool)
        {
            if (options[i].short_name != "")
                short_options += options[i].short_name;
            long_options[i].has_arg = no_argument;
        }
        else
        {
            if (options[i].short_name != "")
                short_options += options[i].short_name + ":";
            long_options[i].has_arg = required_argument;
        }
    }
    long_options[options.size()].name = 0;
    long_options[options.size()].has_arg = 0;
    long_options[options.size()].flag = 0;
    long_options[options.size()].val = 0;

    while (true)
    {
        int index = -1;
        int ch = getopt_long(argc, argv, short_options.c_str(), long_options, &index);

//        if (opterr != 0)
//        {
//            throw exception();
//        }

        if (ch == -1)
            break;

        if (ch == '?')
            throw exception();

        if (ch != 0)
        {
            string s;
            s += ch;

            for (unsigned i = 0; i < options.size(); ++i)
            {
                if (options[i].short_name == s)
                {
                    index = i;
                    break;
                }
            }
        }

        if (options[index].type == OptionBool)
            options[index].value = "true";
        else
            options[index].value = optarg;
    }

    int index = 1;
    for (int i = optind; i < argc; ++i)
    {
        argv[index++] = argv[i];
    }
    argc = index;

    for (unsigned i = 0; i < options.size(); ++i)
    {
        options[i].Process();
    }
}

string OptionDescriptions()
{
    stringstream ss;
    for (unsigned i = 0; i < options.size(); ++i)
    {
        ss << options[i].ToString() << endl;
    }

    return ss.str();
}

bool IsExist(const string &filename)
{
    FILE *fp = fopen(filename.c_str(), "rb");
    if (fp == NULL)
    {
        return false;
    }
    else
    {
        fclose(fp);
        return true;
    }
}

void AddParameter(const char *name, void *pointer, ParameterType type)
{
    parameters[name] = pointer;
    types[name] = type;
}

void ProcessParameters(int &argc, char *argv[])
{
    int k = 0;
    for (int i = 0; i < argc; ++i)
    {
        if (argv[i][0] == '-' && argv[i][1] == '-')
        {
            char *name = argv[i] + 2;
            if (parameters.find(name) != parameters.end())
            {
                switch (types[name])
                {
                    case SIMPLE:
                        *(bool *)parameters[name] = true;
                        break;

                    case INTEGER:
                        *(int *)parameters[name] = atoi(argv[++i]);
                        break;

                    case FLOAT:
                        *(double *)parameters[name] = atof(argv[++i]);
                        break;

                    case STRING:
                        strcpy((char*)parameters[name], argv[++i]);
                }
            }
        }
        else
        {
            argv[k++] = argv[i];
        }
    }
    argv[k] = NULL;
    argc = k;
}


FILE *OpenFile(const char *filename, const char *mode)
{
    FILE *fp = fopen(filename, mode);

    if (fp == NULL)
    {
        LogError("open %s failed\n", filename);
        exit(1);
    }

    return fp;
}

void ReplaceComma(string &s)
{
    for (unsigned i = 0; i < s.size(); ++i)
    {
        if (s[i] == ',')
            s[i] = ' ';
    }
}

void SplitString(const string &s, vector<string> &items)
{
    stringstream ss(s);
    string item;

    while (ss >> item)
        items.push_back(item);
}


//bool ReadFastq(FILE *fp, Sequence *seq, char *comment, char *quality)
//{
//    if (fgets(line, MaxLine, fp) == NULL)
//        return false;
//
//    if (line[0] != '@')
//    {
//        LogError("ReadFastq: comment is not in correct format\n");
//        exit(1);
//    }
//
//    line[strlen(line)-1] = '\0';
//    if (comment != NULL)
//        strcpy(comment, line + 1);
//
//    int offset = 0;
//    while (fgets(line, MaxLine, fp) != NULL)
//    {
//        if (line[0] == '+')
//        {
//            fgets(line, MaxLine, fp);
//            line[strlen(line)-1] = '\0';
//            if (quality != NULL)
//                strcpy(quality, line);
//            break;
//        }
//        else
//        {
//            int len = strlen(line);
//            while (isspace(line[len-1]))
//                --len;
//            copy(line, line + len, buf + offset);
//            offset += len;
//        }
//    }
//
//    //seq->SetContent(buf, offset);
//    buf[offset] = '\0';
//    *seq = buf;
//
//    return offset != 0;
//}
//
//bool ReadFasta(FILE *fp, Sequence *seq, char *comment)
//{
//    if (fgets(line, MaxLine, fp) == NULL)
//        return false;
//
//    if (line[0] != '>')
//    {
//        LogError("ReadFasta: comment is not in correct format\n");
//        exit(1);
//    }
//
//    line[strlen(line)-1] = '\0';
//    if (comment != NULL)
//        strcpy(comment, line + 1);
//
//    int offset = 0;
//    while (fgets(line, MaxLine, fp) != NULL)
//    {
//        if (line[0] == '>')
//        {
//            fseek(fp, -strlen(line), SEEK_CUR);
//            break;
//        }
//        else
//        {
//            int len = strlen(line);
//            while (isspace(line[len-1]))
//                --len;
//            copy(line, line + len, buf + offset);
//            offset += len;
//        }
//    }
//
//    buf[offset] = '\0';
//    //seq->SetContent(buf, offset);
//    *seq = buf;
//    return offset != 0;
//}
//
//void WriteFasta(FILE *fp, Sequence *seq, const char *comment)
//{
//    if (comment != NULL)
//        fprintf(fp, ">%s\n", comment);
//
//    int offset = 0;
//    while (offset < seq->Size())
//    {
//        int lineOffset = offset + FastaMaxLine;
//        if (lineOffset > seq->Size())
//            lineOffset = seq->Size();
//        int ch = (*seq)[lineOffset];
//        (*seq)[lineOffset] = '\0';
//        fprintf(fp, "%s\n", seq->ToCString() + offset);
//        (*seq)[lineOffset] = ch;
//        offset = lineOffset;
//    }
//}
//
//uint64 CountSequences(std::FILE *fp, bool isCheck)
//{
//    uint64 index = 0;
//    while (fgets(line, MaxLine, fp) != NULL)
//    {
//        if (line[0] == '>')
//            ++index;
//    }
//    fseek(fp, 0, SEEK_SET);
//    return index;
//}
//
////uint64 ReadHashGraph(std::FILE *fp, HashGraph *hashGraph, unsigned minCount)
////{
////    hashGraph->Clear();
////    Sequence seq;
////
////    uint64 index = 0;
////    while (ReadFasta(fp, &seq, comment))
////    {
////        unsigned count = 0;
////        for (int i = 0; comment[i]; ++i)
////        {
////            if (comment[i] == '_')
////                comment[i] = ' ';
////        }
////        unsigned in, out;
////        sscanf(comment, "%s %s %u %u %u", buf, buf, &count, &in, &out);
////        seq.Encode();
////
////        if (count >= minCount)
////        {
////            ++index;
////            Kmer kmer = seq.GetKmer(0, seq.Size());
////            HashNode *node = hashGraph->InsertKmer(kmer);
////
////            node->count = count;
////
////            if (node->kmer != kmer)
////                swap(in, out);
////
////            node->in = in;
////            node->out = out;
////        }
////    }
////
////    return index;
////}
////
////uint64 WriteHashGraph(std::FILE *fp, HashGraph *hashGraph, unsigned minCount)
////{
////    uint64 index = 0;
////    Sequence seq;
////
////    for (unsigned i = 0; i < hashGraph->tableSize; ++i)
////    {
////        for (HashNode *node = hashGraph->table[i]; node != NULL; node = node->next)
////        {
////            if (!(node->status & HashGraphFlagDeadend) && node->count >= minCount)
////            {
////                Kmer kmer = node->kmer;
////                seq.SetContent(kmer, kmerLength);
////                seq.Decode();
////                fprintf(fp, ">Kmer%lld_count_%u_%u_%u\n", index++, node->count, node->in, node->out);
////                WriteFasta(fp, &seq);
////            }
////        }
////    }
////
////    return index;
////}
//
//void ReadContigs(FILE *fp, vector<Contig> &contigs, bool code)
//{
//    int total = CountSequences(fp);
//    contigs.resize(total);
//
//    for (int i = 0; i < total; ++i)
//    {
//        ReadFasta(fp, &contigs[i], comment);
//        for (int j = 0; comment[j]; ++j)
//        {
//            if (comment[j] == '_')
//                comment[j] = ' ';
//        }
//
//        sscanf(comment, "%s %s %s %s %s %d", 
//                buf, buf, buf, buf, buf, &contigs[i].Coverage());
//
//        if (code)
//        {
//            contigs[i].Encode();
//        }
//    }
//}
//
//void WriteContigs(FILE *fp, vector<Contig> &contigs, bool code, int minLength)
//{
//    for (int i = 0; i < (int)contigs.size(); ++i)
//    {
//        if (contigs[i].Size() < minLength)
//            continue;
//
//        fprintf(fp, ">contig%d_length%d_coverage%d\n", 
//                i, contigs[i].Size(), contigs[i].Coverage());
//
//        if (code)
//        {
//            contigs[i].Decode();
//            WriteFasta(fp, &contigs[i]);
//            contigs[i].Encode();
//        }
//        else
//            WriteFasta(fp, &contigs[i]);
//    }
//}
//
//void CalculateN50(vector<Contig> &contigs, int minContig, int64 &total, int64 &n50)
//{
//    total = 0;
//    n50 = 0;
//    vector<int> lengths(contigs.size());
//    for (int i = 0; i < (int)lengths.size(); ++i)
//    {
//        lengths[i] = contigs[i].Size();
//        if (lengths[i] >= minContig)
//            total += minContig;
//    }
//    sort(lengths.begin(), lengths.end());
//    reverse(lengths.begin(), lengths.end());
//
//    int64 sum = 0;
//    for (unsigned i = 0; i < lengths.size(); ++i)
//    {
//        sum += lengths[i];
//        if (sum >= 0.5*total && n50 == 0)
//        {
//            n50 = lengths[i];
//            break;
//        }
//    }
//}
//
//// uint64 ReadContigGraph(std::FILE *fp, ContigGraph *contigGraph)
//// {
////     contigGraph->numContigs = CountSequences(fp);
////     contigGraph->contigs = new Sequence[contigGraph->numContigs*2];
////     contigGraph->connections = new vector<Connection>[contigGraph->numContigs*2];
////     contigGraph->paths = new vector<Path>[contigGraph->numContigs*2];
////
////     Sequence seq;
////     for (unsigned i = 0; i < contigGraph->numContigs; ++i)
////     {
////         ReadFasta(fp, &seq);
////         seq.Encode();
////         contigGraph->contigs[i<<1] = seq;
////         seq.ReverseComplement();
////         contigGraph->contigs[(i<<1)|1] = seq;
////     }
////
////     return contigGraph->numContigs;
//// }
