#include "globals.h"
#include "Sequence.h"
#include "Utils.h"
#include "BitOperation.h"
#include "Reader.h"
#include "KmerVector.h"

#include <omp.h>

#include <cstdio>
#include <algorithm>
#include <cstring>
#include <iostream>
#include <cmath>
#include <ctime>
#include <vector>
#include <map>
#include <sstream>


using namespace std;

const int MaxCluster = 2048;

char line[MaxLine];
char buf[MaxLine];
char filename1[MaxLine];
char filename2[MaxLine];
int distance_type = 1;
vector<KmerVector> kmer_vectors[MaxCluster];
int group_num = 5;
int group_size = 1000;
int length = 1000;
int num_classes = 2;

void Usage()
{
    fprintf(stderr, "usage: groupDistanceDistribution [--length l] [--type x] [--groupNum n] [--groupSize s]  pair.list\n");
    fprintf(stderr, "0:median 1:average 2:minimum (default:1)\n");
}

double Distance(KmerVector *kv1, KmerVector *kv2)
{
    double sum = 0;
    for (int i = 0; i < kv1->size; ++i)
        sum += abs(kv1->ranks[i] - kv2->ranks[i]);
    return sum;
}

double DistanceMedian(vector<KmerVector> &kv1, vector<KmerVector> &kv2)
{
    vector<double> tmp(kv1.size() * kv2.size());
#pragma omp parallel for
    for (int i = 0; i < (int)kv1.size(); ++i)
    {
        for (int j = 0; j < (int)kv2.size(); ++j)
        {
            double d = Distance(&kv1[i], &kv2[j]);
            tmp[i*kv2.size() + j] = d;
        }
    }

    sort(tmp.begin(), tmp.end());
    return tmp[tmp.size()/2];
}

double DistanceAverage(vector<KmerVector> &kv1, vector<KmerVector> &kv2)
{
    vector<double> sum(kv1.size());
#pragma omp parallel for
    for (int i = 0; i < (int)kv1.size(); ++i)
    {
        sum[i] = 0;
        for (int j = 0; j < (int)kv2.size(); ++j)
        {
            double d = Distance(&kv1[i], &kv2[j]);
            sum[i] += d;
        }
    }

    double total = 0;
    for (unsigned i = 0; i < sum.size(); ++i)
        total += sum[i];

    int64 count = kv1.size() * kv2.size();
    return total / count;
}

double DistanceMinimum(vector<KmerVector> &kv1, vector<KmerVector> &kv2)
{
    vector<double> minimum(kv1.size());
#pragma omp parallel for
    for (int i = 0; i < (int)kv1.size(); ++i)
    {
        minimum[i] = 1e100;
        for (int j = 0; j < (int)kv2.size(); ++j)
        {
            double d = Distance(&kv1[i], &kv2[j]);
            if (d < minimum[i])
                minimum[i] = d;
        }
    }

    return *min_element(minimum.begin(), minimum.end());
}

void GenGroups(vector<KmerVector> *kmer_vectors, const string &filename)
{
    Sequence ref;
    string comment;
    FastAReader reader(filename);
    reader.Read(ref, comment);
    ref.Encode();

    for (int i = 0; i < group_num; ++i)
    {
        kmer_vectors[i].resize(group_size);
        for (int j = 0; j < group_size; ++j)
        {
            int offset = rand() % (ref.Size() - length + 1);
            Sequence seq;
            ref.GetSubSequence(seq, offset, length);

            kmer_vectors[i][j].Compute(&seq);
            kmer_vectors[i][j].ComputeRank();
        }
    }
}

double GroupDistance(vector<KmerVector> *kmer_vectors1, vector<KmerVector> *kmer_vectors2)
{
    double sum_inter = 0;
    for (int i = 0; i < group_num; ++i)
    {
        for (int j = 0; j < group_num; ++j)
        {
            double d;
            
            if (distance_type == 0)
            {
                d = DistanceMedian(kmer_vectors1[i], kmer_vectors2[j]);
            }
            else if (distance_type == 1)
            {
                d = DistanceAverage(kmer_vectors1[i], kmer_vectors2[j]);
            }
            else
            {
                d = DistanceMinimum(kmer_vectors1[i], kmer_vectors2[j]);
            }

            sum_inter += d;
        }
    }

    return sum_inter /= group_num * group_num;
}

int main(int argc, char *argv[])
{
    AddParameter("classes", &num_classes, INTEGER);
    AddParameter("type", &distance_type, INTEGER);
    AddParameter("groupNum", &group_num, INTEGER);
    AddParameter("groupSize", &group_size, INTEGER);
    AddParameter("length", &length, INTEGER);
    ProcessParameters(argc, argv);

    if (argc < 2)
    {
        Usage();
        exit(1);
    }

    FILE *flist = fopen(argv[1], "rb");
    while (fgets(line, MaxLine, flist) != NULL)
    {
        vector<string> file_list(num_classes);
        stringstream ss(line);
        for (int i = 0; i < num_classes; ++i)
            ss >> file_list[i];

        for (int i = 0; i < num_classes; ++i)
            GenGroups(kmer_vectors + i*group_num, file_list[i]);

        vector<double> intra_dist;
        vector<double> inter_dist;

        for (int i = 0; i < num_classes; ++i)
            intra_dist.push_back(GroupDistance(kmer_vectors + i*group_num, kmer_vectors + i*group_num));
        for (int i = 0; i < num_classes; ++i)
        {
            for (int j = i+1; j < num_classes; ++j)
                inter_dist.push_back(GroupDistance(kmer_vectors + i*group_num, kmer_vectors + j *group_num));
        }

        double intra = *max_element(intra_dist.begin(), intra_dist.end());
        double inter = *min_element(inter_dist.begin(), inter_dist.end());

        int flag = intra < 0.9 * inter;
        //if (intra < 0.7 * inter)
        {
            for (int i = 0; i < num_classes; ++i)
                printf("%s, ", file_list[i].c_str());
            printf("%.2f, %.2f, %d, %.4f\n", intra, inter, flag, intra / inter);
        }

        fflush(NULL);
    }

    return 0;
}

