#include "config.h"

#include "globals.h"
#include "Log.h"
#include "Sequence.h"
#include "SimpleAssembler.h"
#include "IterativeAssembler.h"
#include "ScaffoldAssembler.h"
#include "GappedScaffoldAssembler.h"
#include "IDBAOption.h"
#include "Reader.h"
#include "Read.h"
#include "Writer.h"
#include "Utils.h"

#include <unistd.h>
#include <sys/wait.h>

//#include <boost/program_options.hpp>
//#include <boost/filesystem.hpp>

#include <cstdio>
#include <algorithm>
#include <cstring>
#include <vector>
#include <cmath>
#include <string>

//namespace po = boost::program_options;
//namespace fs = boost::filesystem;

using namespace std;

IDBAOption option;
static Alignment aux_alignments[1000000];

static bool Compare(const Contig &x, const Contig &y)
{
    return x.Size() > y.Size();
}

void BuildKmer()
{
    kmerLength = option.mink;
    vector<Read> reads;
    FastAReader read_reader(option.readfile);

    ((Reader&)read_reader).Read(reads);

    vector<Sequence> long_reads;
    if (option.long_readfile != "")
    {
        FastAReader long_reader(option.long_readfile);
        long_reads.resize(long_reader.NumReads());

        int index = 0;
        Sequence seq;
        string comment;
        while (long_reader.Read(seq, comment))
        {
            seq.Encode();
            long_reads[index++] = seq;
        }
    }

    HashGraph *hash_graph = new HashGraph();

    FILE *fkmer = fopen(option.kmer_file.c_str(), "wb");
    for (int prefix = 0; prefix < (1 << option.prefix_length); ++prefix)
    {
#pragma omp parallel for
        for (int i = 0; i < (int64)reads.size(); ++i)
        {
            if (!reads[i].IsActive() || reads[i].Size() < kmerLength)
                continue;

            Sequence seq;
            seq = reads[i];
            seq.Trim(option.trim);
            hash_graph->InsertSequence(seq, prefix, option.mask);
        }

#pragma omp parallel for
        for (int i = 0; i < (int)long_reads.size(); ++i)
            hash_graph->InsertSequence(long_reads[i], prefix, option.mask);

        cout << "table size " << hash_graph->NumNodes() << " " << hash_graph->Capacity() << endl;
        hash_graph->RefreshVertices(option.min_count);

        hash_graph->Save(fkmer);
        cout << "solid kmer " << hash_graph->NumNodes() << endl;
        hash_graph->Clear();
    }

    delete hash_graph;
    fclose(fkmer);
}

void BuildSimpleGraph()
{
    SimpleAssembler *simple = new SimpleAssembler(option);
    vector<Contig> *contigs = new vector<Contig>();
    simple->Assemble(*contigs);
    delete simple;

    FastAWriter writer(option.graphfile);
    for (unsigned i = 0; i < (*contigs).size(); ++i)
    {
        (*contigs)[i].Decode();
        writer.WriteFormat((*contigs)[i], "contig%d", i);
    }

    delete contigs;
}

void Iterate()
{
    IterativeAssembler simple(option);
    vector<Contig> contigs;
    simple.Assemble(contigs);

    //sort(contigs.begin(), contigs.end(), Compare);

    FastAWriter writer(option.contigfile);
    for (unsigned i = 0; i < contigs.size(); ++i)
    {
        contigs[i].Decode();
        writer.WriteFormat(contigs[i], "contig%d", i);
    }
}

void Scaffold()
{
    kmerLength = option.maxk;
    ScaffoldAssembler simple(option);
    vector<Contig> contigs;
    simple.Assemble(contigs);

    FastAWriter writer(option.scaffile);
    for (unsigned i = 0; i < contigs.size(); ++i)
    {
        contigs[i].Decode();
        writer.WriteFormat(contigs[i], "contig%d", i);
    }
}

void ScaffoldWithGap()
{
    GappedScaffoldAssembler simple(option);
    vector<Contig> contigs;
    simple.Assemble(contigs);

    FastAWriter writer(option.gap_scaffold_file);
    for (unsigned i = 0; i < contigs.size(); ++i)
    {
        contigs[i].Decode();
        writer.WriteFormat(contigs[i], "contig%d", i);
    }
}

void AlignReads()
{
    vector<Contig> contigs;
    kmerLength = option.maxk;
    //scaffold_graph.SetMinPairs(option.min_pairs);
    FastAReader contig_reader(option.contigfile);
    FastAReader read_reader(option.readfile);
    HashAlign aligner;

    Contig contig;
    string comment;
    while (contig_reader.Read(contig, comment))
    {
        int index = comment.size()-1;
        while (isdigit(comment[index]))
            --index;
        contig.sum_coverage = atoi(comment.c_str() + index + 1);
        contig.Encode();
        contigs.push_back(contig);
    }

    cout << "contigs " << contigs.size() << endl;

    sort(contigs.begin(), contigs.end(), Compare);
    //reverse(contigs.begin(), contigs.end());

    //scaffold_graph.Initialize(contigs);

    cout << "scaffold graph initialize" << endl;
    //aligner.Initialize(scaffold_graph.GetContigNodes());
    aligner.Initialize(contigs);
    cout << "aligner initialize" << endl;

    vector<Read> reads;
    Sequence seq1, seq2;
    string comment1, comment2;
    FILE *falign = fopen(option.align_file.c_str(), "wb");
    int BufferSize = 100000;
    vector<vector<Alignment> > alignments(BufferSize);
    //while (read_reader.Read(seq1, comment1) && read_reader.Read(seq2, comment2))
    while (((Reader &)read_reader).Read(reads, BufferSize) != 0)
    {
#pragma omp parallel for
        for (int64 i = 0; i < (int64)reads.size(); ++i)
        {
            Sequence seq = reads[i];
            seq.Trim(option.trim);
            aligner.AlignSequence(&seq, alignments[i]);
        }

        for (unsigned i = 0; i < reads.size(); ++i)
        {
            int size = alignments[i].size();
            copy(alignments[i].begin(), alignments[i].end(), aux_alignments);
            fwrite(&size, sizeof(int), 1, falign);
            fwrite(aux_alignments, sizeof(Alignment), size, falign);
        }
    }
    fclose(falign);
}

void Usage()
{
    cout << "IDBA: Iterative De Bruijn graph short read Assembler" << endl;
    cout << "Version " << VERSION << endl;
    cout << endl;
    cout << "Usage: idba --read read-file [--output out] [options]\n" << endl;
}

int main(int argc, char *argv[])
{
    bool is_help = false;
    AddOption("help", "h", is_help, "produce help message");
    AddOption("read", "r", option.readfile, "read file");
    AddOption("long", "l", option.long_readfile, "long read file");
    AddOption("output", "o", option.prefix, "prefix of output");
    AddOption("scaffold", "", option.is_scaffold, "use pair end information to merge contigs");
    AddOption("mink", "", option.mink, "minimum k value");
    AddOption("maxk", "", option.maxk, "maximum k value");
    AddOption("minCount", "", option.min_count, "filtering threshold for each k-mer");
    AddOption("cover", "", option.cover, "the cutting coverage for contigs");
    AddOption("trim", "", option.trim, "discard n bases at the end of each read");
    AddOption("minPairs", "", option.min_pairs, "minimum number of pair-end connections to join two contigs");
    AddOption("minLength", "", option.min_length, "minimum contig length for output");
    AddOption("prefixLength", "", option.prefix_length, "length of the prefix of k-mer used to split k-mer table");

    try 
    {
        ProcessOptions(argc, argv);
    }
    catch (exception &e)
    {
        Usage();
        cout << "Allowed Options: " << endl;
        cout << OptionDescriptions() << endl;
        exit(1);
    }

    if (is_help || option.readfile == "")
    {
        Usage();
        cout << "Allowed Options: " << endl;
        cout << OptionDescriptions() << endl;
        exit(1);
    }

    option.Compute();
    kmerLength = option.mink;

    if (!IsExist(option.kmer_file))
    {
        int pid = fork();
        if (pid == 0)
        {
            BuildKmer();
            exit(1); 
        }
        else
        {
            int status;
            waitpid(pid, &status, 0);
        }
    }

    if (!IsExist(option.graphfile))
    {
        int pid = fork();
        if (pid == 0)
        {
            BuildSimpleGraph();
            exit(1); 
        }
        else
        {
            int status;
            waitpid(pid, &status, 0);
        }
    }

    if (!IsExist(option.contigfile))
    {
        int pid = fork();
        if (pid == 0)
        {
            Iterate();
            exit(1); 
        }
        else
        {
            int status;
            waitpid(pid, &status, 0);
        }
    }

    if (!IsExist(option.align_file) && option.is_scaffold)
    {
        int pid = fork();
        if (pid == 0)
        {
            //BuildSimpleGraph();
            AlignReads();
            exit(1); 
        }
        else
        {
            int status;
            waitpid(pid, &status, 0);
        }
    }

    if (option.is_scaffold && !IsExist(option.scaffile))
    {
        int pid = fork();
        if (pid == 0)
        {
            Scaffold();
            exit(1); 
        }
        else
        {
            int status;
            waitpid(pid, &status, 0);
        }
    }

    if (option.is_scaffold && option.is_gap && !IsExist(option.gap_scaffold_file))
    {
        int pid = fork();
        if (pid == 0)
        {
            ScaffoldWithGap();
            exit(1); 
        }
        else
        {
            int status;
            waitpid(pid, &status, 0);
        }
    }

    return 0;
}

