#include "config.h"

#include "globals.h"
#include "Log.h"
#include "Sequence.h"
#include "SimpleAssembler.h"
#include "SimpleTranAssembler.h"
#include "IterativeAssembler.h"
#include "ScaffoldAssembler.h"
#include "GappedScaffoldAssembler.h"
#include "IDBAOption.h"
#include "Reader.h"
#include "Writer.h"
#include "Utils.h"

#include <unistd.h>
#include <sys/wait.h>

//#include <boost/program_options.hpp>
//#include <boost/filesystem.hpp>

#include <cstdio>
#include <algorithm>
#include <cstring>
#include <vector>
#include <cmath>
#include <string>

//namespace po = boost::program_options;
//namespace fs = boost::filesystem;

using namespace std;

IDBAOption option;

void BuildSimpleGraph()
{
    SimpleTranAssembler *simple = new SimpleTranAssembler(option);
    //SimpleAssembler *simple = new SimpleAssembler(option);
    vector<Contig> *contigs = new vector<Contig>();
    simple->Assemble(*contigs);
    delete simple;

    FastAWriter writer(option.graphfile);
    for (unsigned i = 0; i < (*contigs).size(); ++i)
    {
        (*contigs)[i].Decode();
        writer.WriteFormat((*contigs)[i], "contig%d", i);
    }

    delete contigs;
}

void Iterate()
{
    IterativeAssembler simple(option);
    vector<Contig> contigs;
    simple.Assemble(contigs);

    FastAWriter writer(option.contigfile);
    for (unsigned i = 0; i < contigs.size(); ++i)
    {
        contigs[i].Decode();
        writer.WriteFormat(contigs[i], "contig%d", i);
    }
}

void Scaffold()
{
    kmerLength = option.maxk;
    ScaffoldAssembler simple(option);
    vector<Contig> contigs;
    simple.Assemble(contigs);

    FastAWriter writer(option.scaffile);
    for (unsigned i = 0; i < contigs.size(); ++i)
    {
        contigs[i].Decode();
        writer.WriteFormat(contigs[i], "contig%d", i);
    }
}

void ScaffoldWithGap()
{
    GappedScaffoldAssembler simple(option);
    vector<Contig> contigs;
    simple.Assemble(contigs);

    FastAWriter writer(option.gap_scaffold_file);
    for (unsigned i = 0; i < contigs.size(); ++i)
    {
        contigs[i].Decode();
        writer.WriteFormat(contigs[i], "contig%d", i);
    }
}

void Usage()
{
    cout << "IDBA: Iterative De Bruijn graph short read Assembler" << endl;
    cout << "Version " << VERSION << endl;
    cout << endl;
    cout << "Usage: idba --read read-file [--output out] [options]\n" << endl;
}

int main(int argc, char *argv[])
{
    bool is_help = false;
    AddOption("help", "h", is_help, "produce help message");
    AddOption("read", "r", option.readfile, "read file");
    AddOption("output", "o", option.prefix, "prefix of output");
    AddOption("scaffold", "", option.is_scaffold, "use pair end information to merge contigs");
    AddOption("mink", "", option.mink, "minimum k value");
    AddOption("maxk", "", option.maxk, "maximum k value");
    AddOption("minCount", "", option.min_count, "filtering threshold for each k-mer");
    AddOption("cover", "", option.cover, "the cutting coverage for contigs");
    AddOption("trim", "", option.trim, "discard n bases at the end of each read");
    AddOption("minPairs", "", option.min_pairs, "minimum number of pair-end connections to join two contigs");
    AddOption("minLength", "", option.min_length, "minimum contig length for output");
    AddOption("prefixLength", "", option.prefix_length, "length of the prefix of k-mer used to split k-mer table");

    ProcessOptions(argc, argv);
//    po::options_description desc("Allowed options");
//    desc.add_options()
//        ("help,h", "produce help message")
//        ("read,r", po::value<string>(&option.readfile), "read file")
//        ("read2", po::value<string>(&option.readfile2), "read file")
//        ("output,o", po::value<string>(&option.prefix)->default_value("out"), "prefix of output")
//        ("scaffold", "use pair end information to merge contigs")
//        ("gap", "use pair end information to merge contigs with gap")
//        ("mink", po::value<int>(&option.mink)->default_value(25), "minimum k value")
//        ("maxk", po::value<int>(&option.maxk)->default_value(50), "maximum k value")
//        ("minCount", po::value<int>(&option.min_count)->default_value(2), "filtering threshold for each k-mer")
//        ("cover", po::value<double>(&option.cover)->default_value(0), "the cutting coverage for contigs")
//        ("trim", po::value<int>(&option.trim)->default_value(0), "discard n bases a the end of each read")
//        ("minPairs", po::value<int>(&option.min_pairs)->default_value(5), "minimum number of pair-end connections to join two contigs")
//        ("minLength", po::value<int>(&option.min_length)->default_value(100), "minimum contig length for output")
//        ("ref", po::value<string>(&option.ref_file), "genome reference");
//    ;
//
//    try
//    {
//        po::variables_map vm;
//        po::store(po::parse_command_line(argc, argv, desc), vm);
//        po::notify(vm);
//
//        option.is_scaffold = vm.count("scaffold");
//        option.is_gap = vm.count("gap");
//
//        if (vm.count("help") || vm.count("read") == 0)
//        {
//            Usage();
//            cout << desc << endl;
//            return 1;
//        }
//    }
//    catch (exception &e)
//    {
//        Usage();
//        cout << desc << endl;
//        return 1;
//    }
    if (is_help || option.readfile == "")
    {
        Usage();
        cout << OptionDescriptions() << endl;
        exit(1);
    }

    option.Compute();

    if (!IsExist(option.graphfile))
    {
        int pid = fork();
        if (pid == 0)
        {
            BuildSimpleGraph();
            exit(1); 
        }
        else
        {
            int status;
            waitpid(pid, &status, 0);
        }
    }

    if (!IsExist(option.contigfile))
    {
        int pid = fork();
        if (pid == 0)
        {
            Iterate();
            exit(1); 
        }
        else
        {
            int status;
            waitpid(pid, &status, 0);
        }
    }

    if (option.is_scaffold && !IsExist(option.scaffile))
    {
        int pid = fork();
        if (pid == 0)
        {
            Scaffold();
            exit(1); 
        }
        else
        {
            int status;
            waitpid(pid, &status, 0);
        }
    }

    if (option.is_scaffold && option.is_gap && !IsExist(option.gap_scaffold_file))
    {
        int pid = fork();
        if (pid == 0)
        {
            ScaffoldWithGap();
            exit(1); 
        }
        else
        {
            int status;
            waitpid(pid, &status, 0);
        }
    }

    return 0;
}

